import os
import sys
import datetime as _dt
from pathlib import Path
from typing import Optional

import polars as pl
import typer

from accmix.scanner import scan_pwm
from accmix.annotate import compute_sl, annotate_tss
from accmix.model_core import fit_model
from accmix import evaluate


app = typer.Typer(help="Accessibility mixture model: scan PWM, annotate sites, and fit/evaluate EM models.")


def _timestamp() -> str:
    """Return a compact timestamp for result file names."""
    return _dt.datetime.now().strftime("%Y%m%d_%H%M%S")


@app.command("scan", help="Scan genome with PWM and compute inner/outer scores.")
def scan_pwm_cmd(
    fasta: Path = typer.Option(..., "-f", "--fasta", help="Genome FASTA (.fa or .fa.gz)."),
    pwm: Path = typer.Option(..., "-p", "--pwm", help="PWM txt with columns Pos,A,C,G,T/U (probabilities)."),
    out_prefix: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out-prefix",
        help="Custom output prefix (default: results/<pwm>.<timestamp>).",
    ),
    M: int = typer.Option(50, "-M", help="Inner half-width in bp (default: 50)."),
    N: int = typer.Option(500, "-N", help="Outer half-width in bp (default: 500; must be >M)."),
    A: int = typer.Option(2_000_000, "-A", help="Top-A entries to keep per strand (default: 2e6)."),
    B: int = typer.Option(2_000_000, "-B", help="Bottom-B entries to keep per strand (default: 2e6)."),
    bg: Optional[str] = typer.Option(
        None,
        "-g",
        "--bg",
        help="Background probs for LLR as 'A,C,G,T' (overridden by --bg-from-fasta).",
    ),
    bg_from_fasta: bool = typer.Option(
        False,
        "-G",
        "--bg-from-fasta",
        help="Estimate background composition from the provided FASTA.",
    ),
) -> None:
    """Scan the genome with a PWM and write strand-aware score tables.

    This reproduces the behavior of the original Snakemake/script interface but
    exposes it as a typed CLI command.
    """

    os.makedirs("results", exist_ok=True)
    pwm_name = pwm.stem
    ts = _timestamp()
    out_prefix_str = str(out_prefix or Path("results") / f"{pwm_name}.{ts}")

    scan_pwm(
        fasta=str(fasta),
        pwm=str(pwm),
        out_prefix=out_prefix_str,
        M=M,
        N=N,
        A=A,
        B=B,
        bg=bg,
        bg_from_fasta=bg_from_fasta,
    )


@app.command("annotate-acc", help="Compute accessibility-derived s_l from AS native/fixed and PWM top table.")
def annotate_acc_cmd(
    ASnative: Path = typer.Option(..., "-n", "--ASnative", help="AS native file contains chr, position, strand, score, depth, and motif information (motif information is not used)."),
    ASfixed: Path = typer.Option(..., "-f", "--ASfixed", help="AS fixed bed6 file contains chr, start, end, strand, score, and depth."),
    toptable: Path = typer.Option(..., "-t", "--toptable", help="PWM scan top-table TSV.gz with inner/outer scores."),
    out: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out",
        help="Output parquet (default: results/sl.<timestamp>.parquet).",
    ),
    M: int = typer.Option(50, "-M", help="Inner half-width used in scanning (default: 50)."),
    N: int = typer.Option(500, "-N", help="Outer half-width used in scanning (default: 500)."),
) -> None:
    """Derive accessibility score s_l around motif sites using AS native/fixed tracks."""

    os.makedirs("results", exist_ok=True)
    out_path = out or Path("results") / f"sl.{_timestamp()}.parquet"
    compute_sl(str(ASnative), str(ASfixed), str(toptable), str(out_path), M=M, N=N)
    typer.echo(f"Wrote {out_path}")


@app.command("annotate-tss", help="Add TSS proximity, conservation, and TPM annotations to sites.")
def annotate_tss_cmd(
    input_parquet: Path = typer.Option(
        ...,
        "-i",
        "--input-parquet",
        help="Input parquet with id and prior features (e.g. output of annotate-acc).",
    ),
    out: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out",
        help="Output parquet (default: results/annotated.<timestamp>.parquet).",
    ),
    rna_seq_parquet: Path = typer.Option(
        Path("data/expression/ENCFF364YCB_HeLa_RNAseq_Transcripts_count_curated.parquet"),
        "-r",
        "--rna-seq-parquet",
        help="RNA-seq transcript-level expression parquet (TPM).",
    ),
    phastcons_bed: Path = typer.Option(
        Path("data/conservation_score/phastCons100way1.bed"),
        "-c",
        "--phastcons-bed",
        help="BED file with phastCons peaks used for regional conservation score.",
    ),
    phastcons_parquet: Path = typer.Option(
        Path("data/conservation_score/hg38.phastCons100way.1.parquet"),
        "-p",
        "--phastcons-parquet",
        help="Parquet with base-level phastCons scores.",
    ),
    phylop_parquet: Path = typer.Option(
        Path("data/conservation_score/hg38.phyloP100way.1.parquet"),
        "-y",
        "--phylop-parquet",
        help="Parquet with base-level phyloP scores.",
    ),
    reference: str = typer.Option(
        "GRCh38",
        "-R",
        "--reference",
        help="Reference genome assembly name (e.g. GRCh38).",
    ),
) -> None:
    """Annotate motif sites with TSS distance, conservation, and expression.

    Warning: this step can be memory-intensive on genome-scale inputs.
    """

    os.makedirs("results", exist_ok=True)
    out_path = out or Path("results") / f"annotated.{_timestamp()}.parquet"
    annotate_tss(
        input_parquet=str(input_parquet),
        out_parquet=str(out_path),
        rna_seq_parquet=str(rna_seq_parquet),
        phastcons_bed=str(phastcons_bed),
        phastcons_parquet=str(phastcons_parquet),
        phylop_parquet=str(phylop_parquet),
        reference=reference,
    )
    typer.echo(f"Wrote {out_path}")


@app.command("model", help="Fit Gaussian EM model with logistic prior on annotated sites.")
def model_cmd(
    input_parquet: Path = typer.Option(
        ...,
        "-i",
        "--input-parquet",
        help="Annotated parquet with site id, s_l column, and feature columns.",
    ),
    out_prefix: Optional[Path] = typer.Option(
        None,
        "-o",
        "--out-prefix",
        help="Output prefix (default: results/<rbp>_<motif>.<timestamp>).",
    ),
    rbp_name: str = typer.Option("RBP", "-r", "--rbp-name", help="RBP name label used in output prefix."),
    motif_id: str = typer.Option("Motif", "-m", "--motif-id", help="Motif ID label used in output prefix."),
    feat_cols: Optional[str] = typer.Option(
        None,
        "-F",
        "--feat-cols",
        help="Comma-separated list of feature column names (otherwise use default set).",
    ),
    site_col: str = typer.Option("id", "-s", "--site-col", help="Site identifier column name."),
    s_col: str = typer.Option("s_l", "-S", "--s-col", help="Accessibility score column name."),
    mode: str = typer.Option(
        "em",
        "-M",
        "--mode",
        help="Modeling mode (currently only 'em' with Gaussian components is supported).",
    ),
) -> None:
    """Fit the EM mixture model and write per-site priors/posteriors and model JSON."""

    if mode != "em":
        typer.echo("Only 'em' mode is currently supported.", err=True)
        typer.Exit(code=1)

    os.makedirs("results", exist_ok=True)
    ts = _timestamp()
    base = out_prefix or Path("results") / f"{rbp_name}_{motif_id}.{ts}"

    feat_list = None
    if feat_cols:
        feat_list = [c.strip() for c in feat_cols.split(",") if c.strip()]

    out_parquet, out_json = fit_model(
        input_parquet=str(input_parquet),
        out_prefix=str(base),
        rbp_name=rbp_name,
        motif_id=motif_id,
        feature_columns=feat_list,
        site_col=site_col,
        s_col=s_col,
    )
    typer.echo(f"Wrote {out_parquet} and {out_json}")


@app.command("evaluate", help="Evaluate fitted models against CLIP/PIP-seq data.")
def evaluate_cmd(
    model_parquet: Path = typer.Option(
        ...,
        "-M",
        "--model-parquet",
        help="Model parquet produced by 'accmix model' (contains prior/posterior).",
    ),
    clipseq_bed: Path = typer.Option(
        ...,
        "-b",
        "--clipseq-bed",
        help="CLIP-seq peaks BED path.",
    ),
    pipseq_parquet: Path = typer.Option(
        ...,
        "-p",
        "--pipseq-parquet",
        help="PIP-seq parquet path.",
    ),
    rbp_name: str = typer.Option("RBP", "-r", "--rbp-name", help="RBP label for titles/filenames."),
    motif_id: str = typer.Option("Motif", "-m", "--motif-id", help="Motif ID for titles/filenames."),
    output_root: Path = typer.Option(Path("results"), "-o", "--output-root", help="Directory to write plots/logs."),
    motif_logo: Optional[Path] = typer.Option(None, "-L", "--motif-logo", help="Optional motif logo image."),
    score_phastcons100_threshold: float = typer.Option(1.0, "-t", "--score-phastcons100-threshold", help="Conservation score threshold."),
    motif_range: int = typer.Option(50, "-R", "--motif-range", help="Half-window around site position for overlaps."),
) -> None:
    """Run the evaluation pipeline using direct CLI options."""

    evaluate.run_evaluation(
        model_parquet=str(model_parquet),
        clipseq_bed=str(clipseq_bed),
        pipseq_parquet=str(pipseq_parquet),
        rbp_name=rbp_name,
        motif_id=motif_id,
        output_root=str(output_root),
        motif_logo=str(motif_logo) if motif_logo else None,
        score_phastcons100_threshold=score_phastcons100_threshold,
        motif_range=motif_range,
    )


def main() -> None:
    """Entry point for the `accmix` command."""
    app()



"""Memory-optimized annotate variant that keeps wide tables inside Polars."""

from __future__ import annotations
import argparse
from typing import Optional
import os
import polars as pl
import pyranges as pr
import math
import tempfile
import shutil
from pathlib import Path

def compute_sl(ASnative: str, ASfixed: str, toptable: str, out_parquet: str,
               M: int = 50, N: int = 500, threshold: int = 8) -> None:
    AS_fixed = pl.read_csv(ASfixed, has_header=False,
                           new_columns=["Chromosome", "Start", "Strand", "AS_fixed", 'depth_fixed', "motif"],
                           separator="\t")\
                   .select(["Chromosome", "Start", "Strand", "AS_fixed", 'depth_fixed'])\
                   .with_columns((pl.col("Start") + 1).alias("End"))
    AS_native = pl.read_csv(ASnative, has_header=False,
                            new_columns=["Chromosome", "Start", "Strand", "AS_native", 'depth_native', "motif"],
                            separator="\t")\
                    .select(["Chromosome", "Start", "Strand", "AS_native", 'depth_native'])\
                    .with_columns((pl.col("Start") + 1).alias("End"))
    AS_fixed_pr = pr.PyRanges(AS_fixed.to_pandas())
    AS_native_pr = pr.PyRanges(AS_native.to_pandas())
    dftop = pl.read_csv(toptable, separator="\t")\
                .with_columns((pl.col("inner_mean_logPWM") - pl.col("outer_mean_logPWM")).alias("diff_logPWM"))\
                .sort(['chrom', 'pos1', 'strand'])\
                .with_columns(pl.concat_str(['chrom', 'pos1', 'strand'], separator="_").alias("id"))

    dftop_inner = dftop.with_columns((pl.col("pos1") - pl.lit(M)).alias("Start"), (pl.col("pos1") + pl.lit(M)).alias("End"))\
                       .rename({"chrom": "Chromosome", "strand": "Strand"})
    dftop_outer = dftop.with_columns((pl.col("pos1") - pl.lit(N)).alias("Start"), (pl.col("pos1") + pl.lit(N)).alias("End"))\
                       .rename({"chrom": "Chromosome", "strand": "Strand"})
    dftop_inner_pr = pr.PyRanges(dftop_inner.to_pandas())
    dftop_outer_pr = pr.PyRanges(dftop_outer.to_pandas())

    dftop_inner_fixed_pr = dftop_inner_pr.join_overlaps(AS_fixed_pr, strand_behavior="same", join_type = "inner")
    dftop_outer_fixed_pr = dftop_outer_pr.join_overlaps(AS_fixed_pr, strand_behavior="same", join_type = "inner")
    dftop_inner_native_pr = dftop_inner_pr.join_overlaps(AS_native_pr, strand_behavior="same", join_type = "inner")
    dftop_outer_native_pr = dftop_outer_pr.join_overlaps(AS_native_pr, strand_behavior="same", join_type = "inner")


    dftop_inner_fixed_df = pl.from_pandas(dftop_inner_fixed_pr)
    dftop_inner_native_df = pl.from_pandas(dftop_inner_native_pr)

    def filter_outer(df: pl.DataFrame) -> pl.DataFrame:
        flank = max(N - M, 0)
        mask1 = (df["Start_b"] >= df["Start"]) & (df["Start_b"] <= df["Start"] + flank)
        mask2 = (df["Start_b"] >= df["End"] - flank) & (df["Start_b"] <= df["End"])  
        return df.filter(mask1 | mask2)

    dftop_outer_fixed_df = filter_outer(pl.from_pandas(dftop_outer_fixed_pr))
    dftop_outer_native_df = filter_outer(pl.from_pandas(dftop_outer_native_pr))

    def summarize_weighted_by_id(
        df: pl.DataFrame,
        value_col: str,
        weight_col: str,
        *,
        id_col: str = "id",
        threshold: int = threshold,
        ddof: int = 0,
    ) -> pl.DataFrame:
        val = pl.col(value_col)
        w = pl.col(weight_col)
        w_mean_name = f"{value_col}_w_mean"
        res = (
            df
            .group_by(id_col)
            .agg([
                pl.len().alias("n"),
                val.mean().alias(f"{value_col}_mean"),
                val.var(ddof=ddof).alias(f"{value_col}_var"),
                (val * w).sum().alias("_wx"),
                w.sum().alias("_w"),
                (w * val.pow(2)).sum().alias("_wx2"),
                pl.first("inner_mean_logPWM"),
                pl.first("outer_mean_logPWM"),
                pl.first("GC_inner_pct"),
                pl.first("GC_outer_pct"),
                pl.first("kmer_count_inner"),
                pl.first("kmer_count_outer"),
                pl.first("diff_logPWM"),
            ])
            .filter(pl.col("n") >= threshold)
            .with_columns([
                (pl.col("_wx") / pl.col("_w")).alias(w_mean_name),
            ])
            .with_columns(
                (pl.col("_wx2") / pl.col("_w") - pl.col(w_mean_name).pow(2)).alias(f"{value_col}_w_var"),
            )
            .drop(['_wx', '_wx2'])
        )
        return res

    dftop_outer_fixed_sum = summarize_weighted_by_id(dftop_outer_fixed_df, value_col="AS_fixed", weight_col="depth_fixed")
    dftop_outer_native_sum = summarize_weighted_by_id(dftop_outer_native_df, value_col="AS_native", weight_col="depth_native")
    dftop_inner_fixed_sum = summarize_weighted_by_id(dftop_inner_fixed_df, value_col="AS_fixed", weight_col="depth_fixed")
    dftop_inner_native_sum = summarize_weighted_by_id(dftop_inner_native_df, value_col="AS_native", weight_col="depth_native")

    pwm_columns = ['inner_mean_logPWM', 'outer_mean_logPWM', 'GC_inner_pct', 'GC_outer_pct', 'kmer_count_inner', 'kmer_count_outer', 'diff_logPWM']

    fixed_part = dftop_outer_fixed_sum.join(dftop_inner_fixed_sum, on="id", how="inner", suffix="_inner")\
                           .drop([i+"_inner" for i in pwm_columns])\
                           .with_columns(
                               ((pl.col("AS_fixed_mean_inner") - pl.col("AS_fixed_mean")) /
                                (pl.col("AS_fixed_var_inner")/pl.col("_w_inner") + pl.col("AS_fixed_var")/pl.col("_w"))).alias("t_fixed")
                           )
    native_part = dftop_outer_native_sum.join(dftop_inner_native_sum, on="id", how="inner", suffix="_inner")\
                           .drop([i+"_inner" for i in pwm_columns])\
                           .with_columns(
                               ((pl.col("AS_native_mean_inner") - pl.col("AS_native_mean")) /
                                (pl.col("AS_native_var_inner")/pl.col("_w_inner") + pl.col("AS_native_var")/pl.col("_w"))).alias("t_native")
                           )

    results_df = fixed_part.join(native_part, on="id", suffix="_native")\
                            .with_columns((pl.col("t_native").pow(2) + pl.col("t_fixed").pow(2)).alias("s_l"))
    # Sort sites by genomic order (chrom, position, strand) before writing
    results_df = (
        results_df
        .with_columns([
            pl.col("id").str.split("_").alias("_id_split")
        ])
        .with_columns([
            pl.col("_id_split").list.get(0).alias("_chrom"),
            pl.col("_id_split").list.get(1).cast(pl.Int64).alias("_pos"),
            pl.col("_id_split").list.get(2).alias("_strand"),
        ])
        .sort(["_chrom", "_pos", "_strand"])  # preserves within-chrom numeric order
        .drop(["_id_split", "_chrom", "_pos", "_strand"])  # clean up temps
    )
    results_df.write_parquet(out_parquet)
def annotate_tss(
    input_parquet: str,
    out_parquet: str,
    rna_seq_parquet: str = "data/expression/ENCFF364YCB_HeLa_RNAseq_Transcripts_count_curated.parquet",
    phastcons_bed: str = "data/conservation_score/phastCons100way1.bed",
    phastcons_parquet: str = "data/conservation_score/hg38.phastCons100way.1.parquet",
    phylop_parquet: str = "data/conservation_score/hg38.phyloP100way.1.parquet",
    reference: str = "GRCh38",
    *,
    # Optional BigWig fast path (if provided, skips Parquet scanning for scores)
    phastcons_bigwig: Optional[str] = None,
    phylop_bigwig: Optional[str] = None,
    # Streaming tile size and buffer for Parquet path
    tile_mb: int = 5,
    window_buffer: int = 100_000,
):
    """Memory-optimized TSS annotation that scales to all chromosomes.

    - If BigWig files are provided, fetch per-site conservation scores (fastest, lowest memory).
    - Otherwise, stream by chromosome in tiles (tile_mb) and scan conservation Parquet only for
      the current window ± buffer, writing chunk outputs and combining with a final deterministic sort.
    """
    from metagene import load_reference, map_to_transcripts
    # Helpers
    def _drop_if_exists(df: pl.DataFrame | pl.LazyFrame, cols: list[str]):
        if isinstance(df, pl.LazyFrame):
            # For LazyFrame, drop ignores missing columns gracefully
            return df.drop([c for c in cols])
        existing = [c for c in cols if c in df.columns]
        return df.drop(existing) if existing else df

    # Load static resources once
    RNA_seq_data = pl.read_parquet(rna_seq_parquet)
    conservative_df = pl.read_csv(
        phastcons_bed, separator="\t", has_header=False,
        new_columns=["Chromosome", "Start", "End", "Name"],
        comment_prefix="#",
    ).with_columns(pl.col("Chromosome").str.replace('chr', ''))
    conservative_pr = pr.PyRanges(conservative_df.to_pandas())
    reference_df = load_reference(reference)

    # Load all sites and parse IDs (IDENTICAL to original)
    results_df = pl.read_parquet(input_parquet)
    ann_df = results_df.with_columns([
        pl.col("id").str.split("_").alias("id_split")
    ]).with_columns([
        pl.col("id_split").list.get(0).str.replace('chr', '').alias("Chromosome"),
        pl.col("id_split").list.get(2).alias("Strand"),
        pl.col("id_split").list.get(1).cast(pl.Int64).alias("Start")
    ]).drop("id_split").with_columns((pl.col("Start") + 50).alias("End"))\
      .with_columns((pl.col("Start") - 50).alias("Start"))

    # Map to transcripts (IDENTICAL to original)
    ann_df_tx = map_to_transcripts(ann_df, reference_df)
    ann_df_tx = ann_df_tx.with_columns([
        (1 / (1 + pl.col("transcript_start")/1000)).fill_null(0).alias("TSS_proximity"),
        (1 / (1 + pl.col("transcript_start") / pl.col("transcript_length"))).fill_null(0).alias("Tx_percent")
    ])
    
    # PyRanges overlap (IDENTICAL to original)
    ann_df_tx_pr = pr.PyRanges(ann_df_tx.to_pandas())
    joined_pr = ann_df_tx_pr.join_overlaps(conservative_pr, strand_behavior="ignore", join_type="left")
    joined_ann_df = pl.DataFrame(joined_pr)
    
    # Compute PhastCons100_percent and join RNA-seq (IDENTICAL to original)
    joined_ann_df = (
        joined_ann_df
        .with_columns(
            (
                ((pl.min_horizontal("End", "End_b") - pl.max_horizontal("Start", "Start_b")) / 100)
                .clip(lower_bound=0)
                .alias("PhastCons100_percent")
            )
        )
        .join(RNA_seq_data, on="transcript_id", how="left")
        .with_columns(pl.col("TPM").fill_null(0))
    )

    # Prepare chrom and start for conservation join
    joined_ann_df = joined_ann_df.with_columns(
        pl.concat_str(pl.lit("chr"), pl.col("Chromosome")).alias("chrom"), 
        (pl.col("Start") + 50).alias("start")
    )

    # OPTIMIZATION: Process per chromosome to bound memory
    use_bigwig = phastcons_bigwig is not None and phylop_bigwig is not None
    chrom_results = []
    
    for chrom in joined_ann_df["chrom"].unique().sort():
        chrom_sites = joined_ann_df.filter(pl.col("chrom") == chrom)
        if chrom_sites.height == 0:
            continue
        
        if use_bigwig:
            # BigWig: vectorized interval fetching per chromosome
            try:
                import pyBigWig  # type: ignore
                import numpy as np
                if 'bw_phast' not in locals():
                    bw_phast = pyBigWig.open(phastcons_bigwig)
                    bw_phyl = pyBigWig.open(phylop_bigwig)
                
                positions = chrom_sites["start"].to_numpy()
                vals_pc = []
                vals_pl = []
                
                # Fetch in batches for better cache locality
                for pos in positions:
                    try:
                        pc = bw_phast.values(chrom, int(pos), int(pos)+1, numpy=True)
                        pl_v = bw_phyl.values(chrom, int(pos), int(pos)+1, numpy=True)
                        vals_pc.append(float(pc[0]) if pc is not None and len(pc) > 0 and not np.isnan(pc[0]) else None)
                        vals_pl.append(float(pl_v[0]) if pl_v is not None and len(pl_v) > 0 and not np.isnan(pl_v[0]) else None)
                    except Exception:
                        vals_pc.append(None)
                        vals_pl.append(None)
                
                chrom_result = chrom_sites.with_columns([
                    pl.Series("score_phastcons100", vals_pc),
                    pl.Series("score_phylop100", vals_pl),
                ])
                chrom_results.append(chrom_result)
            except Exception as e:
                raise RuntimeError(f"BigWig error on {chrom}: {e}")
        else:
            # Parquet: scan only this chromosome's conservation data
            min_pos = max(0, chrom_sites["start"].min() - window_buffer)
            max_pos = chrom_sites["start"].max() + window_buffer
            
            phast_chrom = (
                pl.scan_parquet(phastcons_parquet)
                .filter(
                    (pl.col("chrom") == chrom) & 
                    (pl.col("start") >= min_pos) & 
                    (pl.col("start") <= max_pos)
                )
            )
            phyl_chrom = (
                pl.scan_parquet(phylop_parquet)
                .filter(
                    (pl.col("chrom") == chrom) & 
                    (pl.col("start") >= min_pos) & 
                    (pl.col("start") <= max_pos)
                )
            )
            
            # Join for this chromosome only
            chrom_result = (
                chrom_sites.lazy()
                .sort("start")
                .join_asof(phast_chrom, on="start", strategy="backward", suffix="_phastcons")
                .filter(pl.col("start") <= pl.col("end"))
                .join_asof(phyl_chrom, on="start", strategy="backward", suffix="_phylop100")
                .filter(pl.col("start") <= pl.col("end_phylop100"))
                .rename({"score": "score_phastcons100"})
                .collect(streaming=True)
            )
            chrom_results.append(chrom_result)
    
    # Close BigWig files if used
    if use_bigwig:
        try:
            bw_phast.close()
            bw_phyl.close()
        except Exception:
            pass
    
    # Combine all chromosomes
    joined_ann_joined_df = pl.concat(chrom_results).lazy()
    
    # Clean up columns (IDENTICAL to original)
    joined_ann_joined_df = (
        joined_ann_joined_df if isinstance(joined_ann_joined_df, pl.LazyFrame) else joined_ann_joined_df.lazy()
    ).drop([
        "Chromosome", "Start", "End", "Strand", "gene_id", "transcript_id", 
        "transcript_start", "transcript_end", "transcript_length", 
        'start_codon_pos', 'stop_codon_pos', 'exon_number', 'record_id', 
        'Start_exon', 'End_exon', 'Start_b', 'End_b', 'Name', 
        'end', 'end_phylop100', 'start', 'chrom',
        'chrom_phastcons', 'chrom_phylop100'  # Extra columns from join_asof suffixes
    ])
    
    # Add missing score rename if from BigWig path
    if use_bigwig:
        joined_ann_joined_df = joined_ann_joined_df.rename({"score": "score_phylop100"})

    # Final genomic sort (IDENTICAL to original)
    joined_ann_joined_df = (
        joined_ann_joined_df
        .with_columns([
            pl.col("id").str.split("_").alias("_id_split")
        ])
        .with_columns([
            pl.col("_id_split").list.get(0).alias("_chrom"),
            pl.col("_id_split").list.get(1).cast(pl.Int64).alias("_pos"),
            pl.col("_id_split").list.get(2).alias("_strand"),
        ])
        .sort(["_chrom", "_pos", "_strand"])
        .drop(["_id_split", "_chrom", "_pos", "_strand"])
    )
    joined_ann_joined_df.sink_parquet(out_parquet)


def annotate(
    input_parquet: Optional[str] = None,
    ASnative: Optional[str] = None,
    ASfixed: Optional[str] = None,
    toptable: Optional[str] = None,
    out_parquet: Optional[str] = None,
):
    os.makedirs("results", exist_ok=True)
    if input_parquet is None:
        if not (ASnative and ASfixed and toptable):
            raise SystemExit("Need input_parquet OR all of ASnative/ASfixed/toptable")
        tmp = (out_parquet or "results/annotated.parquet").replace("annotated", "sl")
        compute_sl(ASnative, ASfixed, toptable, tmp)
        input_parquet = tmp
    annotate_tss(input_parquet, out_parquet or "results/annotated.parquet")


def build_arg_parser() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(prog="accmix annotate")
    # annotate-acc parameters
    ap.add_argument("--ASnative", default=None, help="Accessibility native bed6 path")
    ap.add_argument("--ASfixed", default=None, help="Accessibility fixed bed6 path")
    ap.add_argument("--toptable", default=None, help="TopA PWM prior tsv.gz from accmix scan")
    ap.add_argument("--M", type=int, default=50, help="Inner half-width (default 50)")
    ap.add_argument("--N", type=int, default=500, help="Outer half-width (default 500)")
    ap.add_argument("--threshold", type=int, default=8, help="Minimum overlaps per id (default 8)")
    # annotate-tss parameters
    ap.add_argument("--input-parquet", default=None, help="Parquet to annotate (e.g., output of annotate-acc)")
    ap.add_argument("--out", default=None, help="Output parquet path for annotated features")
    ap.add_argument("--rna-seq-parquet", default="data/expression/ENCFF364YCB_HeLa_RNAseq_Transcripts_count_curated.parquet")
    ap.add_argument("--phastcons-bed", default="data/conservation_score/phastCons100way1.bed")
    ap.add_argument("--phastcons-parquet", default="data/conservation_score/hg38.phastCons100way.1.parquet")
    ap.add_argument("--phylop-parquet", default="data/conservation_score/hg38.phyloP100way.1.parquet")
    ap.add_argument("--phastcons-bigwig", default=None, help="Optional BigWig for phastCons values (fast, low memory)")
    ap.add_argument("--phylop-bigwig", default=None, help="Optional BigWig for phyloP values (fast, low memory)")
    ap.add_argument("--tile-mb", type=int, default=5, help="Tile size in megabases for streaming Parquet scan")
    ap.add_argument("--window-buffer", type=int, default=100000, help="Extra bases to include around each tile")
    ap.add_argument("--reference", default="GRCh38")
    return ap



from __future__ import annotations

import os
from typing import Dict, Any, List, Tuple

import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
import numpy as np
import polars as pl
from scipy import stats
from sklearn.metrics import roc_curve, auc, precision_recall_curve, average_precision_score


def compute_qq_validate_distribution(
    out_df: pl.DataFrame,
    variable: str = "s_l",
    addition: float = 1.0,
    save_figure_path: str = "plots/distribution.png",
) -> Dict[str, Dict[str, Dict[str, Any]]]:
    """Distribution diagnostics for log-transformed variable by source label.

    Returns a nested dict of summary statistics per source and distribution
    family (Gamma, Normal, Student-t).
    """

    sources = [s for s in out_df["source"].unique().to_list() if s != "non_determined"]
    distributions = [("Gamma", stats.gamma), ("Normal", stats.norm), ("Student-t", stats.t)]

    fig, axes = plt.subplots(len(sources), 3, figsize=(8, 2 * len(sources)))
    if len(sources) == 1:
        axes = axes.reshape(1, -1)

    results: Dict[str, Dict[str, Dict[str, Any]]] = {}

    for i, source in enumerate(sources):
        log_data = np.log(out_df.filter(pl.col("source") == source)[variable].to_numpy() + addition)
        results[source] = {}

        for j, (name, dist) in enumerate(distributions):
            if name == "Gamma":
                params = dist.fit(log_data, floc=0)
            else:
                params = dist.fit(log_data)

            stats.probplot(log_data, dist=dist, sparams=params, plot=axes[i, j])

            theoretical = axes[i, j].get_lines()[1].get_xdata()
            empirical = axes[i, j].get_lines()[1].get_ydata()
            r2 = float(np.corrcoef(theoretical, empirical)[0, 1] ** 2)
            ks_stat, ks_p = stats.kstest(log_data, lambda x: dist.cdf(x, *params))
            qq_deviation = float(np.mean(np.abs(empirical - theoretical)))

            axes[i, j].set_title(f"{source} - {name}\nR²={r2:.3f}, QQ-dev={qq_deviation:.3f}\nKS p={ks_p:.3f}")
            axes[i, j].grid(True, alpha=0.3)

            results[source][name] = {
                "r2": r2,
                "ks_p": float(ks_p),
                "qq_deviation": qq_deviation,
                "params": params,
                "n_samples": len(log_data),
            }

    os.makedirs(os.path.dirname(save_figure_path) or ".", exist_ok=True)
    plt.tight_layout()
    plt.savefig(save_figure_path)
    plt.close(fig)
    return results


def heatmap_plot(
    data: pl.DataFrame,
    columns: List[str],
    no_normalize_columns: List[str],
    save_heatmap_path: str,
    file_title: str,
    motif_logo: str | None = None,
) -> None:
    """Plot feature heatmap ordered by posterior_r and source label."""

    label_mapping = {
        "TSS_proximity": "TSS Proximity",
        "score_phastcons100": "PhastCons100way score",
        "score_phylop100": "PhyloP100way score",
        "PhastCons100_percent": "PhastCons100way region score",
        "inner_mean_logPWM": "PWM score: inner region",
        "outer_mean_logPWM": "PWM score: outer region",
        "GC_inner_pct": "GC content: inner region",
        "GC_outer_pct": "GC content: outer region",
        "s_l": "RAD-seq signal",
        "prior_p": "Prior",
        "posterior_r": "Posterior",
        "source": "CLIP label",
    }

    special_transforms = {"s_l": lambda x: np.log(x + 1)}

    pdf = data.select(columns).to_pandas()

    def zscore_to_01(x: np.ndarray) -> np.ndarray:
        z = stats.zscore(x)
        return (z - np.nanmin(z)) / (np.nanmax(z) - np.nanmin(z) + 1e-12)

    for col in columns[:-1]:
        if col in no_normalize_columns:
            pdf[f"{col}_norm"] = pdf[col]
            continue
        if col in special_transforms:
            pdf[col] = special_transforms[col](pdf[col])
        pdf[f"{col}_norm"] = zscore_to_01(pdf[col].to_numpy())

    pdf["source_num"] = pdf["source"].map({"clip_bound": 1.0, "non_determined": 0.5, "clip_unbound": 0.0})
    pdf_sorted = pdf.sort_values(["posterior_r", "source_num"])

    desired_order = [
        "TSS_proximity_norm",
        "PhastCons100_percent_norm",
        "score_phastcons100_norm",
        "score_phylop100_norm",
        "inner_mean_logPWM_norm",
        "outer_mean_logPWM_norm",
        "s_l_norm",
        "prior_p_norm",
        "posterior_r_norm",
        "source_num",
    ]

    norm_cols = [f"{col}_norm" for col in columns[:-1]] + ["source_num"]
    norm_cols = [c for c in desired_order if c in norm_cols]
    matrix = pdf_sorted[norm_cols].values.T

    cmap = plt.matplotlib.colors.LinearSegmentedColormap.from_list(
        "wb",
        [(0, "white"), (0.3, "white"), (0.7, "lightblue"), (1, "blue")],
    )
    fig, ax = plt.subplots(figsize=(7.5, 3.5))
    im = ax.imshow(matrix, cmap=cmap, aspect="auto", interpolation="none", rasterized=True)

    ax.set_xlabel("Sample")
    ax.set_title(file_title, pad=20)
    ax.set_yticks(range(len(norm_cols)))

    professional_labels: List[str] = []
    for col in norm_cols:
        original_col = col.replace("_norm", "").replace("_num", "")
        professional_labels.append(label_mapping.get(original_col, original_col))
    ax.set_yticklabels(professional_labels)
    plt.colorbar(im, ax=ax, shrink=0.6)

    if motif_logo and os.path.exists(motif_logo):
        try:
            logo_img = mpimg.imread(motif_logo)
            imagebox = OffsetImage(logo_img, zoom=0.2)
            ab = AnnotationBbox(
                imagebox,
                (0.85, 1.15),
                xycoords="axes fraction",
                frameon=False,
                box_alignment=(0.5, 0.5),
            )
            ax.add_artist(ab)
        except Exception:
            pass

    os.makedirs(os.path.dirname(save_heatmap_path) or ".", exist_ok=True)
    plt.tight_layout()
    plt.subplots_adjust(top=0.85)
    plt.savefig(save_heatmap_path, dpi=300, bbox_inches="tight")
    plt.close(fig)


def plot_em_auc(out_df: pl.DataFrame, file_title: str, save_path: str | None = None) -> float:
    """ROC AUC for EM posterior_r vs CLIP-bound / clip_unbound labels."""

    eval_df = out_df.filter(pl.col("source") != "non_determined").select(["source", "posterior_r"]).to_pandas()
    y_true = (eval_df["source"] == "clip_bound").astype(int).to_numpy()
    y_scores = eval_df["posterior_r"].to_numpy()

    fpr, tpr, _ = roc_curve(y_true, y_scores)
    roc_auc = auc(fpr, tpr)

    fig, ax = plt.subplots(figsize=(3, 3))
    ax.plot(fpr, tpr, color="darkorange", lw=2, label=f"ROC (AUC = {roc_auc:.3f})")
    ax.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="--", label="Random")
    ax.set_xlim(0.0, 1.0)
    ax.set_ylim(0.0, 1.05)
    ax.set_xlabel("False Positive Rate")
    ax.set_ylabel("True Positive Rate")
    ax.set_title(file_title)
    ax.legend(loc="lower right")
    ax.grid(True, alpha=0.3)
    plt.tight_layout()

    if save_path:
        os.makedirs(os.path.dirname(save_path) or ".", exist_ok=True)
        plt.savefig(save_path, dpi=300, bbox_inches="tight")
    plt.close(fig)
    return float(roc_auc)


def plot_em_auc_control(out_df: pl.DataFrame) -> float:
    """Baseline AUC using inner_mean_logPWM as the score."""

    eval_df = (
        out_df.filter(pl.col("source") != "non_determined")
        .select(["source", "inner_mean_logPWM"])
        .to_pandas()
    )
    y_true = (eval_df["source"] == "clip_bound").astype(int).to_numpy()
    y_scores = eval_df["inner_mean_logPWM"].to_numpy()

    fpr, tpr, _ = roc_curve(y_true, y_scores)
    return float(auc(fpr, tpr))


def change_label(out_df: pl.DataFrame) -> pl.DataFrame:
    """Flip posterior/prior labels if conservation suggests negative selection.

    If PhastCons-based difference is negative while PhyloP is not strongly
    positive, posterior_r and prior_p are inverted.
    """

    diff_phylop = (
        out_df.select(
            (
                pl.when(pl.col("posterior_r") > 0.9)
                .then(pl.col("score_phylop100"))
                .mean()
                - pl.when(pl.col("posterior_r") < 0.1)
                .then(pl.col("score_phylop100"))
                .mean()
            ).alias("mean_diff")
        ).item()
    )
    diff_phastcons = (
        out_df.select(
            (
                pl.when(pl.col("posterior_r") > 0.9)
                .then(pl.col("score_phastcons100"))
                .mean()
                - pl.when(pl.col("posterior_r") < 0.1)
                .then(pl.col("score_phastcons100"))
                .mean()
            ).alias("mean_diff")
        ).item()
    )

    if (diff_phylop > 0) and (diff_phastcons > 0):
        return out_df
    if diff_phastcons < 0:
        return out_df.with_columns(
            (1 - pl.col("posterior_r")).alias("posterior_r"),
            (1 - pl.col("prior_p")).alias("prior_p"),
        )
    return out_df


def save_analysis_results(
    model: Dict[str, Any] | None,
    source_coefs: np.ndarray,
    source_coefs_s_l: np.ndarray,
    source_intercept: float,
    source_intercept_s_l: float,
    feature_columns: List[str],
    roc_auc: float,
    roc_auc_control: float,
    auc_source: float,
    auc_source_s_l: float,
    file_title: str,
    out_df: pl.DataFrame,
    test_dict: Dict[str, Any] | None,
    motif_df: pl.DataFrame | None = None,
    output_dir: str = "logs",
) -> Tuple[pl.DataFrame, pl.DataFrame, pl.DataFrame]:
    """Persist beta coefficients and key evaluation metrics to TSVs."""

    os.makedirs(output_dir, exist_ok=True)

    def safe_counts(df_pl: pl.DataFrame | None) -> Tuple[int, int, int]:
        if df_pl is None:
            return 0, 0, 0
        vc = df_pl["source"].value_counts().to_pandas().set_index("source")
        def get(k: str) -> int:
            return int(vc.loc[k, "count"]) if k in vc.index else 0
        return get("clip_bound"), get("clip_unbound"), get("non_determined")

    def best_f1(y_true: np.ndarray, y_score: np.ndarray) -> Tuple[float, float]:
        prec, rec, thr = precision_recall_curve(y_true, y_score)
        f1_all = 2 * prec[:-1] * rec[:-1] / (prec[:-1] + rec[:-1] + 1e-12)
        if len(f1_all) == 0 or np.all(np.isnan(f1_all)):
            return 0.0, 0.5
        idx = int(np.nanargmax(f1_all))
        return float(f1_all[idx]), float(thr[idx])

    def precision_at_fpr(
        y_true: np.ndarray, y_score: np.ndarray, target_fpr: float
    ) -> Tuple[float, float, float, float]:
        fpr, tpr, thr = roc_curve(y_true, y_score)
        if len(fpr) == 0:
            return 0.0, 0.0, 0.0, 0.5
        valid = fpr <= target_fpr
        if valid.any():
            sub_idx = np.argmax(tpr[valid])
            idx = np.flatnonzero(valid)[sub_idx]
        else:
            idx = int(np.argmin(np.abs(fpr - target_fpr)))
        thr_sel = float(thr[idx])
        y_pred = (y_score >= thr_sel).astype(int)
        p = int(((y_pred == 1) & (y_true == 1)).sum())
        fp = int(((y_pred == 1) & (y_true == 0)).sum())
        prec = float(p / (p + fp + 1e-12))
        return prec, float(fpr[idx]), float(tpr[idx]), thr_sel

    if model is not None and isinstance(model, dict) and "beta" in model:
        beta_em = model["beta"]
    else:
        # Fallback: fill with NaNs when model params are unavailable
        beta_em = [float("nan")] * (len(feature_columns) + 1)
    beta_lr = np.concatenate([[source_intercept], source_coefs])
    beta_lr_s_l = np.concatenate([[source_intercept_s_l], source_coefs_s_l])

    site_counts = out_df["source"].value_counts().to_pandas().set_index("source")
    n_clip_bound = int(site_counts.loc["clip_bound", "count"]) if "clip_bound" in site_counts.index else 0
    n_clip_unbound = int(site_counts.loc["clip_unbound", "count"]) if "clip_unbound" in site_counts.index else 0
    n_non_determined = int(site_counts.loc["non_determined", "count"]) if "non_determined" in site_counts.index else 0

    m_clip_bound, m_clip_unbound, m_non_determined = safe_counts(motif_df)

    filt = out_df.filter(pl.col("source") != "non_determined")
    if filt.height > 0:
        y_true = (filt["source"] == "clip_bound").to_numpy().astype(int)
        y_score = filt["posterior_r"].to_numpy()
        try:
            ap_em = float(average_precision_score(y_true, y_score))
        except Exception:
            ap_em = float("nan")
        try:
            f1_best, f1_thr = best_f1(y_true, y_score)
        except Exception:
            f1_best, f1_thr = float("nan"), float("nan")
        try:
            prec_10_fpr, act_fpr_10, act_tpr_10, thr_10 = precision_at_fpr(y_true, y_score, 0.10)
        except Exception:
            prec_10_fpr = act_fpr_10 = act_tpr_10 = thr_10 = float("nan")
        try:
            prec_1_fpr, act_fpr_1, act_tpr_1, thr_1 = precision_at_fpr(y_true, y_score, 0.01)
        except Exception:
            prec_1_fpr = act_fpr_1 = act_tpr_1 = thr_1 = float("nan")
    else:
        ap_em = f1_best = f1_thr = float("nan")
        prec_10_fpr = act_fpr_10 = act_tpr_10 = thr_10 = float("nan")
        prec_1_fpr = act_fpr_1 = act_tpr_1 = thr_1 = float("nan")

    qq_stats = {
        "clip_bound": {},
        "clip_unbound": {},
    }
    if test_dict is not None:
        for src in ["clip_bound", "clip_unbound"]:
            for dist_name in ["Gamma", "Normal", "Student-t"]:
                qq_stats[src][dist_name] = (
                    test_dict.get(src, {})
                    .get(dist_name, {})
                    .get("qq_deviation", None)
                )

    beta_results_df = pl.DataFrame(
        {
            "feature": ["intercept"] + feature_columns,
            "beta_em": beta_em,
            "beta_lr": beta_lr,
        }
    )
    beta_results_df_s_l = pl.DataFrame(
        {
            "feature": ["intercept", "s_l"] + feature_columns,
            "beta_lr_s_l": beta_lr_s_l,
        }
    )

    metadata_df = pl.DataFrame(
        {
            "file_title": [file_title],
            "auc_control": [roc_auc_control],
            "auc_em": [roc_auc],
            "ap_em": [ap_em],
            "f1_best_em": [f1_best],
            "f1_best_threshold_em": [f1_thr],
            "precision_at_10pct_fpr_em": [prec_10_fpr],
            "actual_fpr_at_10pct_target": [act_fpr_10],
            "tpr_at_10pct_fpr_em": [act_tpr_10],
            "threshold_at_10pct_fpr_em": [thr_10],
            "precision_at_1pct_fpr_em": [prec_1_fpr],
            "actual_fpr_at_1pct_target": [act_fpr_1],
            "tpr_at_1pct_fpr_em": [act_tpr_1],
            "threshold_at_1pct_fpr_em": [thr_1],
            "auc_lr": [auc_source],
            "auc_lr_sl": [auc_source_s_l],
            "n_clip_bound_out_df": [n_clip_bound],
            "n_clip_unbound_out_df": [n_clip_unbound],
            "n_non_determined_out_df": [n_non_determined],
            "n_clip_bound_motif_df": [m_clip_bound],
            "n_clip_unbound_motif_df": [m_clip_unbound],
            "n_non_determined_motif_df": [m_non_determined],
            "qq_clip_bound_gamma": [qq_stats["clip_bound"].get("Gamma")],
            "qq_clip_bound_normal": [qq_stats["clip_bound"].get("Normal")],
            "qq_clip_bound_student": [qq_stats["clip_bound"].get("Student-t")],
            "qq_clip_unbound_gamma": [qq_stats["clip_unbound"].get("Gamma")],
            "qq_clip_unbound_normal": [qq_stats["clip_unbound"].get("Normal")],
            "qq_clip_unbound_student": [qq_stats["clip_unbound"].get("Student-t")],
        }
    )

    beta_results_df_s_l.write_csv(os.path.join(output_dir, f"{file_title}_beta_coefficients_s_l.tsv"), separator="\t")
    beta_results_df.write_csv(os.path.join(output_dir, f"{file_title}_beta_coefficients.tsv"), separator="\t")
    metadata_df.write_csv(os.path.join(output_dir, f"{file_title}_metadata.tsv"), separator="\t")

    return beta_results_df_s_l, beta_results_df, metadata_df


def plot_precision_recall_em(
    out_df: pl.DataFrame, file_title: str, save_path: str | None = None
) -> float:
    """Precision–recall curve and average precision for EM posteriors."""

    filtered_df = out_df.filter(pl.col("source") != "non_determined")
    y_true = (filtered_df["source"] == "clip_bound").to_numpy().astype(int)
    y_scores = filtered_df["posterior_r"].to_numpy()

    precision, recall, _ = precision_recall_curve(y_true, y_scores)
    avg_precision = average_precision_score(y_true, y_scores)
    baseline = float(np.sum(y_true) / len(y_true)) if len(y_true) else 0.0

    fig, ax = plt.subplots(figsize=(3, 3))
    ax.plot(recall, precision, "b-", linewidth=2, label=f"Model (AP = {avg_precision:.3f})")
    ax.axhline(y=baseline, color="r", linestyle="--", label=f"Random (AP = {baseline:.3f})")
    ax.set_xlim(0.0, 1.0)
    ax.set_ylim(0.0, 1.05)
    ax.set_xlabel("Recall")
    ax.set_ylabel("Precision")
    ax.set_title(f"Precision–Recall Curve\n{file_title}")
    ax.legend(loc="lower left")
    ax.grid(True, alpha=0.3)
    plt.tight_layout()

    if save_path:
        os.makedirs(os.path.dirname(save_path) or ".", exist_ok=True)
        plt.savefig(save_path, dpi=300, bbox_inches="tight")
    plt.close(fig)
    return float(avg_precision)



from __future__ import annotations
import argparse, gzip, sys, heapq
from typing import Optional, Tuple, List, Iterable
import numpy as np

# --- FASTA I/O ---
def open_maybe_gzip(path: str):
    return gzip.open(path, "rt") if path.endswith(".gz") else open(path, "rt")

def fasta_records(path: str) -> Iterable[Tuple[str, str]]:
    with open_maybe_gzip(path) as fh:
        name = None
        buf = []
        for line in fh:
            if not line:
                break
            if line[0] == '>':
                if name is not None:
                    yield name, ''.join(buf).upper()
                name = line[1:].strip().split()[0]
                buf = []
            else:
                buf.append(line.strip())
        if name is not None:
            yield name, ''.join(buf).upper()

# --- PWM & background ---
def read_pwm_txt(path: str) -> np.ndarray:
    with open(path, "r") as fh:
        lines = [ln.strip() for ln in fh if ln.strip()]
    header = lines[0].split()
    col = {h.upper(): i for i, h in enumerate(header)}
    if "U" in col and "T" in col:
        raise ValueError("PWM header cannot contain both U and T.")
    if "U" in col:
        col["T"] = col["U"]
    need = ["POS", "A", "C", "G", "T"]
    for k in need:
        if k not in col:
            raise ValueError(f"Missing column '{k}' in PWM header: {header}")
    data = []
    for ln in lines[1:]:
        toks = ln.split()
        A = float(toks[col["A"]]); C = float(toks[col["C"]])
        G = float(toks[col["G"]]); T = float(toks[col["T"]])
        data.append([A, C, G, T])
    pwm = np.array(data, dtype=np.float64)
    eps = 1e-9
    pwm = pwm + eps
    pwm = pwm / pwm.sum(axis=1, keepdims=True)
    return np.log(pwm)

def parse_bg_arg(s: str) -> np.ndarray:
    vals = [float(x) for x in s.split(",")]
    if len(vals) != 4:
        raise ValueError("Expected 4 comma-separated probs for --bg (A,C,G,T).")
    arr = np.array(vals, dtype=np.float64)
    if not np.isfinite(arr).all() or (arr <= 0).any():
        raise ValueError("Background probs must be positive & finite.")
    arr /= arr.sum()
    return arr

def estimate_bg_from_fasta(fasta_path: str) -> np.ndarray:
    cnt = np.zeros(4, dtype=np.int64)
    for _, seq in fasta_records(fasta_path):
        a = np.frombuffer(seq.encode('ascii'), dtype=np.uint8)
        cnt[0] += np.count_nonzero(a == 65)
        cnt[1] += np.count_nonzero(a == 67)
        cnt[2] += np.count_nonzero(a == 71)
        cnt[3] += np.count_nonzero((a == 84) | (a == 85))
    q = cnt.astype(np.float64)
    q = np.where(q > 0, q, 1.0)
    q /= q.sum()
    return q

def apply_background_llr(logp_rows: np.ndarray, q: Optional[np.ndarray]) -> np.ndarray:
    if q is None:
        return logp_rows
    logq = np.log(q)
    return logp_rows - logq[None, :]

# --- Optional Numba ---
try:
    from numba import njit
    NUMBA_OK = True
except Exception:
    NUMBA_OK = False

def build_kmer_table_fwd(logp: np.ndarray) -> np.ndarray:
    L = logp.shape[0]
    K = 4 ** L
    fwd = np.empty(K, dtype=np.float32)
    for code in range(K):
        s = 0.0
        c = code
        for pos in range(L):
            b = c & 3
            s += logp[L - 1 - pos, b]
            c >>= 2
        fwd[code] = s
    return fwd

if NUMBA_OK:
    @njit(cache=True, fastmath=True)
    def score_kmers_forward(seq_codes: np.ndarray, L: int, table_fwd: np.ndarray):
        n = seq_codes.size
        T = n - L + 1
        sf = np.zeros(T, dtype=np.float32)
        valid = np.zeros(T, dtype=np.uint8)
        if T <= 0:
            return sf, valid
        mask = (1 << (2*L)) - 1
        code = 0
        valid_len = 0
        for i in range(n):
            b = seq_codes[i]
            if b < 4:
                code = ((code << 2) & mask) | b
                valid_len += 1
            else:
                code = 0
                valid_len = 0
            if valid_len >= L:
                t = i - L + 1
                sf[t] = table_fwd[code]
                valid[t] = 1
        return sf, valid

    @njit(cache=True, fastmath=True)
    def cumsum_int8_to_int32(x: np.ndarray) -> np.ndarray:
        out = np.empty(x.size + 1, dtype=np.int32)
        s = 0
        out[0] = 0
        for i in range(x.size):
            s += int(x[i])
            out[i+1] = s
        return out

    @njit(cache=True, fastmath=True)
    def cumsum_f32_to_f64(x: np.ndarray) -> np.ndarray:
        out = np.empty(x.size + 1, dtype=np.float64)
        s = 0.0
        out[0] = 0.0
        for i in range(x.size):
            s += float(x[i])
            out[i+1] = s
        return out

def base_to_code_array(seq: str) -> np.ndarray:
    m = np.full(len(seq), 255, dtype=np.uint8)
    a = np.frombuffer(seq.encode('ascii'), dtype=np.uint8, count=len(seq))
    m[(a == 65)] = 0
    m[(a == 67)] = 1
    m[(a == 71)] = 2
    m[(a == 84) | (a == 85)] = 3
    return m

def gc_bool_array(seq: str) -> np.ndarray:
    a = np.frombuffer(seq.encode('ascii'), dtype=np.uint8, count=len(seq))
    return ((a == 67) | (a == 71)).astype(np.uint8)

def reverse_complement(seq: str) -> str:
    seq = seq.replace('U', 'T').replace('u', 't')
    return seq.translate(str.maketrans("ACGTacgt", "TGCAtgca"))[::-1]

def fallback_score_forward(seq: str, logp: np.ndarray):
    L = logp.shape[0]
    n = len(seq)
    T = n - L + 1
    sf = np.zeros(max(T, 0), dtype=np.float32)
    valid = np.zeros(max(T, 0), dtype=np.uint8)
    for t in range(T):
        kmer = seq[t:t+L]
        if any(c not in "ACGT" for c in kmer):
            continue
        s1 = 0.0
        for k, ch in enumerate(kmer):
            b = "ACGT".index(ch)
            s1 += logp[k, b]
        sf[t] = s1
        valid[t] = 1
    return sf, valid

def aggregate_regions_on_sequence(seq: str, name: str, strand: str,
                                  logp: np.ndarray, fwd_table: np.ndarray,
                                  M: int, N: int,
                                  topA_heap: List[Tuple[float, tuple]],
                                  botB_heap: List[Tuple[float, tuple]],
                                  A: int, B: int):
    L = logp.shape[0]
    n = len(seq)
    if n < max(2*M+1, 2*N+1, L):
        return

    if NUMBA_OK:
        sf, valid = score_kmers_forward(base_to_code_array(seq), L, fwd_table)
    else:
        sf, valid = fallback_score_forward(seq, logp)

    gc_arr = gc_bool_array(seq)

    T = n - L + 1
    w_in  = max(0, (2*M + 1) - L + 1)
    w_out = max(0, (N - M) - L + 1)

    block = 5_000_000
    for i0 in range(0, n, block):
        i1 = min(n, i0 + block)
        idx = np.arange(i0, i1, dtype=np.int64)

        t_end_in = idx + (M - L + 1)
        t_end_ol = idx - (M + L)
        t_end_or = idx + (N - L + 1)

        # Inner
        t_min_in = max(0, int(t_end_in.min() - (w_in - 1)))
        t_max_in = min(T - 1, int(t_end_in.max()))
        if w_in > 0 and t_min_in <= t_max_in:
            sl_in = sf[t_min_in:t_max_in+1]
            vl_in = valid[t_min_in:t_max_in+1]
            ps_in = cumsum_f32_to_f64(sl_in) if NUMBA_OK else np.concatenate([[0.0], sl_in.cumsum(dtype=np.float64)])
            pc_in = cumsum_int8_to_int32(vl_in) if NUMBA_OK else np.concatenate([[0], vl_in.cumsum(dtype=np.int32)])
            te_in = (t_end_in - t_min_in).clip(min=0, max=t_max_in - t_min_in)
            sum_in = ps_in[te_in + 1] - ps_in[np.maximum(te_in - (w_in - 1), 0)]
            cnt_in = pc_in[te_in + 1] - pc_in[np.maximum(te_in - (w_in - 1), 0)]
        else:
            sum_in = np.zeros(idx.size, dtype=np.float64)
            cnt_in = np.zeros(idx.size, dtype=np.int32)

        # Outer left
        t_min_ol = max(0, int(t_end_ol.min() - (w_out - 1)))
        t_max_ol = min(T - 1, int(t_end_ol.max()))
        if w_out > 0 and t_min_ol <= t_max_ol:
            sl_ol = sf[t_min_ol:t_max_ol+1]
            vl_ol = valid[t_min_ol:t_max_ol+1]
            ps_ol = cumsum_f32_to_f64(sl_ol) if NUMBA_OK else np.concatenate([[0.0], sl_ol.cumsum(dtype=np.float64)])
            pc_ol = cumsum_int8_to_int32(vl_ol) if NUMBA_OK else np.concatenate([[0], vl_ol.cumsum(dtype=np.int32)])
            te_ol = (t_end_ol - t_min_ol).clip(min=0, max=t_max_ol - t_min_ol)
            sum_ol = ps_ol[te_ol + 1] - ps_ol[np.maximum(te_ol - (w_out - 1), 0)]
            cnt_ol = pc_ol[te_ol + 1] - pc_ol[np.maximum(te_ol - (w_out - 1), 0)]
        else:
            sum_ol = np.zeros(idx.size, dtype=np.float64)
            cnt_ol = np.zeros(idx.size, dtype=np.int32)

        # Outer right
        t_min_or = max(0, int(t_end_or.min() - (w_out - 1)))
        t_max_or = min(T - 1, int(t_end_or.max()))
        if w_out > 0 and t_min_or <= t_max_or:
            sl_or = sf[t_min_or:t_max_or+1]
            vl_or = valid[t_min_or:t_max_or+1]
            ps_or = cumsum_f32_to_f64(sl_or) if NUMBA_OK else np.concatenate([[0.0], sl_or.cumsum(dtype=np.float64)])
            pc_or = cumsum_int8_to_int32(vl_or) if NUMBA_OK else np.concatenate([[0], vl_or.cumsum(dtype=np.int32)])
            te_or = (t_end_or - t_min_or).clip(min=0, max=t_max_or - t_min_or)
            sum_or = ps_or[te_or + 1] - ps_or[np.maximum(te_or - (w_out - 1), 0)]
            cnt_or = pc_or[te_or + 1] - pc_or[np.maximum(te_or - (w_out - 1), 0)]
        else:
            sum_or = np.zeros(idx.size, dtype=np.float64)
            cnt_or = np.zeros(idx.size, dtype=np.int32)

        with np.errstate(divide="ignore", invalid="ignore"):
            mean_in = np.where(cnt_in > 0, sum_in / cnt_in, np.nan)
            sum_o = sum_ol + sum_or
            cnt_o = cnt_ol + cnt_or
            mean_o = np.where(cnt_o > 0, sum_o / cnt_o, np.nan)

        a0 = max(0, i0 - N)
        a1 = min(n, i1 + N)
        gc_slice = gc_arr[a0:a1]
        gc_ps = np.concatenate([[0], gc_slice.cumsum(dtype=np.int32)])

        L_in = (np.minimum(idx + M, n-1) - np.maximum(idx - M, 0) + 1).astype(np.int32)
        s_in = np.maximum(idx - M, 0) - a0
        e_in = np.minimum(idx + M, n-1) - a0
        gc_in = (gc_ps[e_in + 1] - gc_ps[s_in]).astype(np.int32)
        gc_in_pct = np.where(L_in > 0, gc_in * 100.0 / L_in, np.nan)

        s_ol = np.maximum(idx - N, 0) - a0
        e_ol = np.maximum(idx - M - 1, -1) - a0
        s_or = np.minimum(idx + M + 1, n) - a0
        e_or = np.minimum(idx + N, n-1) - a0
        e_ol_clip = np.maximum(e_ol, -1)
        len_ol = (e_ol_clip - s_ol + 1).clip(min=0)
        len_or = (e_or - s_or + 1).clip(min=0)
        gc_ol = np.where(len_ol > 0, gc_ps[e_ol_clip + 1] - gc_ps[s_ol], 0)
        gc_or = np.where(len_or > 0, gc_ps[e_or + 1] - gc_ps[s_or], 0)
        len_o = len_ol + len_or
        gc_o = gc_ol + gc_or
        gc_o_pct = np.where(len_o > 0, gc_o * 100.0 / len_o, np.nan)

        pos1 = idx + 1
        for j in range(idx.size):
            if cnt_in[j] <= 0:
                continue
            rec = (
                name,
                int(pos1[j]),
                strand,
                float(mean_in[j]),
                float(mean_o[j]) if not np.isnan(mean_o[j]) else float("nan"),
                float(gc_in_pct[j]), float(gc_o_pct[j]),
                int(cnt_in[j]), int(cnt_o[j])
            )
            score = float(mean_in[j])
            if len(topA_heap) < A:
                heapq.heappush(topA_heap, (score, rec))
            else:
                if score > topA_heap[0][0]:
                    heapq.heapreplace(topA_heap, (score, rec))
            neg = -score
            if len(botB_heap) < B:
                heapq.heappush(botB_heap, (neg, rec))
            else:
                if neg > botB_heap[0][0]:
                    heapq.heapreplace(botB_heap, (neg, rec))

def write_tsv_gz_stream(path: str, row_iter: Iterable[Tuple]):
    with gzip.open(path, "wt") as gz:
        gz.write(
            "chrom\tpos1\tstrand\tinner_mean_logPWM\touter_mean_logPWM\t"
            "GC_inner_pct\tGC_outer_pct\tkmer_count_inner\tkmer_count_outer\n"
        )
        for r in row_iter:
            gz.write("\t".join(map(str, r)) + "\n")

def scan_pwm(
    fasta: str,
    pwm: str,
    out_prefix: Optional[str] = None,
    M: int = 50,
    N: int = 500,
    A: int = 2_000_000,
    B: int = 2_000_000,
    bg: Optional[str] = None,
    bg_from_fasta: bool = False,
):
    if out_prefix is None:
        raise ValueError("out_prefix is required")
    if N <= M:
        raise ValueError("Require N > M so that outer flanks are non-empty.")

    logp = read_pwm_txt(pwm)
    bg_q = None
    if bg_from_fasta:
        bg_q = estimate_bg_from_fasta(fasta)
    elif bg is not None:
        bg_q = parse_bg_arg(bg)
    logp = apply_background_llr(logp, bg_q)

    fwd_table = build_kmer_table_fwd(logp)
    topA_heap: List[Tuple[float, tuple]] = []
    botB_heap: List[Tuple[float, tuple]] = []

    contig_len = {}
    contig_order = {}
    for name, seq_plus in fasta_records(fasta):
        contig_len[name] = len(seq_plus)
        if name not in contig_order:
            contig_order[name] = len(contig_order)
        sys.stderr.write(f"[info] Processing {name} len={len(seq_plus):,} L={logp.shape[0]} M={M} N={N}\n")
        aggregate_regions_on_sequence(seq_plus, name, '+', logp, fwd_table,
                                      M, N, topA_heap, botB_heap, A, B)
        seq_minus = reverse_complement(seq_plus)
        aggregate_regions_on_sequence(seq_minus, name, '-', logp, fwd_table,
                                      M, N, topA_heap, botB_heap, A, B)

    topA = sorted(topA_heap, key=lambda x: -x[0])
    botB = sorted(botB_heap, key=lambda x: x[0])

    def remap_rows(rows: List[Tuple]):
        out = []
        for _, rec in rows:
            chrom, pos1o, strand, inner_mean, outer_mean, gc_in, gc_out, kc_in, kc_out = rec
            if strand == '-':
                n = contig_len[chrom]
                pos1 = n - pos1o + 1
            else:
                pos1 = pos1o
            out.append((chrom, pos1, strand, inner_mean, outer_mean, gc_in, gc_out, kc_in, kc_out))
        return out

    topA_rows = remap_rows(topA)
    botB_rows = remap_rows(botB)

    # Partition rows per chromosome to avoid one huge sort; then stream in FASTA order.
    def per_chrom_sorted_rows(rows: List[Tuple]):
        buckets = {}
        for r in rows:
            buckets.setdefault(r[0], []).append(r)
        for chrom, order in sorted(contig_order.items(), key=lambda x: x[1]):
            if chrom in buckets:
                lst = buckets[chrom]
                lst.sort(key=lambda x: x[1])  # sort by pos1 within chrom
                for rec in lst:
                    yield rec

    outA = f"{out_prefix}_topA.tsv.gz"
    outB = f"{out_prefix}_botB.tsv.gz"
    write_tsv_gz_stream(outA, per_chrom_sorted_rows(topA_rows))
    write_tsv_gz_stream(outB, per_chrom_sorted_rows(botB_rows))
    sys.stderr.write(f"[done] Wrote {outA} and {outB}\n")


def build_arg_parser() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(prog="accmix scan")
    ap.add_argument("--fasta", required=True)
    ap.add_argument("--pwm", required=True)
    ap.add_argument("--out-prefix", default=None)
    ap.add_argument("-M", type=int, default=50)
    ap.add_argument("-N", type=int, default=500)
    ap.add_argument("-A", type=int, default=2_000_000)
    ap.add_argument("-B", type=int, default=2_000_000)
    ap.add_argument("--bg", default=None)
    ap.add_argument("--bg-from-fasta", action="store_true")
    return ap



from __future__ import annotations
from pathlib import Path
from typing import List, Optional

import numpy as np
import pyranges as pr
import polars as pl
import pandas as pd
from scipy import stats
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import roc_curve, auc

from accmix.model_core import read_polars_input
from accmix import utils as utils


def run_evaluation(
    model_parquet: str,
    clipseq_bed: str,
    pipseq_parquet: str,
    rbp_name: str,
    motif_id: str,
    output_root: str,
    motif_logo: Optional[str] = None,
    score_phastcons100_threshold: float = 1.0,
    motif_range: int = 50,
) -> None:
    """Run evaluation pipeline.

    Parameters
    - model_parquet: Parquet produced by `accmix model` containing prior/posterior.
    - clipseq_bed: CLIP-seq peaks BED path.
    - pipseq_parquet: PIP-seq parquet path.
    - rbp_name: RBP label for titles/filenames.
    - motif_id: Motif identifier for titles/filenames.
    - output_root: Directory to write plots and logs.
    - motif_logo: Optional logo to display on heatmaps.
    - score_phastcons100_threshold: Threshold for class labeling logic.
    - motif_range: Window half-size around site position for overlap checks.
    """

    output_root = Path(output_root)

    output_root.mkdir(parents=True, exist_ok=True)

    file_title = f"{rbp_name}_{motif_id}"

    df = pl.read_parquet(model_parquet)
    missing_cols = [c for c in ("prior_p", "posterior_r", "id", "s_l") if c not in df.columns]
    if missing_cols:
        raise ValueError(f"Model parquet missing required columns: {missing_cols}")

    # Read CLIP-seq with pandas (more robust to odd encodings), then convert to Polars
    clipseq_pd = pd.read_csv(
        clipseq_bed,
        sep="\t",
        header=None,
        names=[
            "Chromosome",
            "Start",
            "End",
            "peak_id",
            "score",
            "Strand",
            "RBP_Name",
            "method",
            "cell_line",
            "datasource",
        ],
    )
    clipseq_df = pl.from_pandas(clipseq_pd)

    df = df.with_columns([
        pl.col("id").str.split("_").alias("id_split")
    ]).with_columns([
        pl.col("id_split").list.get(0).alias("Chromosome"),
        pl.col("id_split").list.get(2).alias("Strand"),
        pl.col("id_split").list.get(1).cast(pl.Int64).alias("Position"),
    ]).drop("id_split")

    df = df.with_columns([
        (pl.col("Position") - motif_range).alias("Start"),
        (pl.col("Position") + motif_range).alias("End"),
    ])

    motif_pr = pr.PyRanges(df.to_pandas())
    clipseq_pr = pr.PyRanges(
        clipseq_df.select([
            "Chromosome",
            "Start",
            "End",
            "Strand",
            "peak_id",
            "score",
        ]).to_pandas()
    )

    pipseq_pr = pr.PyRanges(pl.read_parquet(pipseq_parquet).to_pandas())

    # pyranges >= 0.3 uses `overlap` instead of `intersect`
    intersections = motif_pr.overlap(clipseq_pr, strand_behavior="same")
    pipintersections = motif_pr.overlap(pipseq_pr, strand_behavior="same")

    intersected_ids = set(intersections["id"].tolist()) if len(intersections) > 0 else set()
    pipintersected_ids = set(pipintersections["id"].tolist()) if len(pipintersections) > 0 else set()

    df = df.with_columns([
        pl.when(pl.col("id").is_in(list(intersected_ids)))
        .then(pl.lit("clip_bound"))
        .when(
            ~pl.col("id").is_in(list(pipintersected_ids))
            & (pl.col("score_phastcons100") <= score_phastcons100_threshold)
        )
        .then(pl.lit("clip_unbound"))
        .otherwise(pl.lit("non_determined"))
        .alias("source"),
    ])

    df = df.drop(["Chromosome", "Position", "Start", "End"])

    feature_columns: List[str] = [
        "inner_mean_logPWM",
        "outer_mean_logPWM",
        "GC_inner_pct",
        "GC_outer_pct",
        "TSS_proximity",
        "PhastCons100_percent",
        "TPM",
        "score_phastcons100",
        "score_phylop100",
    ]
    skip_normalization = [
        "TSS_proximity",
        "PhastCons100_percent",
        "score_phastcons100",
    ]

    df_pl, s, X, site_ids = read_polars_input(df, "id", "s_l", feature_columns)
    X_zscore = X.copy()
    for i, col in enumerate(feature_columns):
        if col not in skip_normalization:
            col_zscore = stats.zscore(X[:, i])
            X_zscore[:, i] = np.nan_to_num(col_zscore, nan=0.0)
    X_zscore[:, 0] = 1
    log_s = np.log(s + 1)

    out_df = df_pl
    out_df = utils.change_label(out_df)

    roc_auc = utils.plot_em_auc(out_df, file_title, save_path=None)
    roc_auc_control = utils.plot_em_auc_control(out_df)

    expanded_feature_columns = feature_columns + [
        "prior_p",
        "posterior_r",
        "s_l",
        "source",
    ]

    plots_dir = output_root / "plots"
    logs_dir = output_root / "logs"
    plots_dir.mkdir(parents=True, exist_ok=True)
    logs_dir.mkdir(parents=True, exist_ok=True)

    utils.heatmap_plot(
        data=out_df,
        columns=expanded_feature_columns,
        no_normalize_columns=skip_normalization + ["source"],
        save_heatmap_path=str(plots_dir / f"{file_title}_heatmap_with_title_logo.png"),
        file_title=file_title,
        motif_logo=motif_logo,
    )

    dist_dir = plots_dir / "distribution"
    dist_dir.mkdir(parents=True, exist_ok=True)
    test_dict = utils.compute_qq_validate_distribution(
        out_df,
        save_figure_path=str(dist_dir / f"{file_title}.dist.val.png"),
    )

    source_mask = out_df["source"] != "non_determined"
    X_source = X_zscore[source_mask.to_numpy(), 1:]
    y_source_labels = out_df.filter(source_mask)["source"].to_numpy()
    y_source = (y_source_labels == "clip_bound").astype(int)

    lr_source = LogisticRegression(max_iter=1000, random_state=42)
    lr_source.fit(X_source, y_source)
    y_source_proba = lr_source.predict_proba(X_source)[:, 1]
    fpr_source, tpr_source, _ = roc_curve(y_source, y_source_proba)
    auc_source = auc(fpr_source, tpr_source)

    X_zscore_s_l = X_zscore.copy()
    X_zscore_s_l[:, 0] = stats.zscore(log_s)
    X_source_s_l = X_zscore_s_l[source_mask.to_numpy(), :]
    lr_source_s_l = LogisticRegression(max_iter=1000, random_state=42)
    lr_source_s_l.fit(X_source_s_l, y_source)
    y_source_proba_s_l = lr_source_s_l.predict_proba(X_source_s_l)[:, 1]
    fpr_source_s_l, tpr_source_s_l, _ = roc_curve(y_source, y_source_proba_s_l)
    auc_source_s_l = auc(fpr_source_s_l, tpr_source_s_l)

    beta_results_df_s_l, beta_results_df, metadata_df = utils.save_analysis_results(
        None,
        lr_source.coef_[0],
        lr_source_s_l.coef_[0],
        lr_source.intercept_[0],
        lr_source_s_l.intercept_[0],
        feature_columns,
        roc_auc,
        roc_auc_control,
        auc_source,
        auc_source_s_l,
        file_title,
        out_df,
        test_dict=test_dict,
        motif_df=df,
        output_dir=str(logs_dir),
    )

    print("Evaluation completed for", file_title)



from __future__ import annotations
import argparse
import json
from typing import Optional, Tuple, Dict, List
import polars as pl
import numpy as np
from scipy import stats

# ---- Internalized pieces from scripts/model.py ----
def softclip(x: np.ndarray, lo: float = -40.0, hi: float = 40.0) -> np.ndarray:
    return np.clip(x, lo, hi)

def sigmoid(eta: np.ndarray) -> np.ndarray:
    return 1.0 / (1.0 + np.exp(-softclip(eta)))

def add_intercept(X: np.ndarray) -> np.ndarray:
    return np.column_stack([np.ones((X.shape[0], 1)), X])

def irls_logistic(X: np.ndarray, r: np.ndarray, max_iter: int = 100, tol: float = 1e-6) -> np.ndarray:
    n, p = X.shape
    beta = np.zeros(p)
    for _ in range(max_iter):
        eta = X @ beta
        p_hat = sigmoid(eta)
        w = np.clip(p_hat * (1 - p_hat), 1e-9, None)
        z = eta + (r - p_hat) / np.clip(w, 1e-12, None)
        WX = X * w[:, None]
        H = X.T @ WX
        g = X.T @ (w * z)
        try:
            beta_new = np.linalg.solve(H, g)
        except np.linalg.LinAlgError:
            beta_new = np.linalg.solve(H + 1e-6 * np.eye(p), g)
        if np.linalg.norm(beta_new - beta) < tol * (1.0 + np.linalg.norm(beta)):
            beta = beta_new
            break
        beta = beta_new
    return beta

def read_polars_input(df: pl.DataFrame, site_col: str, s_col: str, feat_cols: List[str]) -> Tuple[pl.DataFrame, np.ndarray, np.ndarray, np.ndarray]:
    required = {site_col, s_col, *feat_cols}
    missing = [c for c in required if c not in df.columns]
    if missing:
        raise ValueError(f"Missing required columns: {missing}")
    required_cols = [site_col, s_col] + feat_cols
    df = df.drop_nulls(subset=required_cols)
    s = df[s_col].to_numpy()
    finite_mask = np.isfinite(s)
    if not np.all(finite_mask):
        df = df.filter(pl.Series("finite_mask", finite_mask))
        s = s[finite_mask]
    X = df.select(feat_cols).to_numpy()
    X = add_intercept(X)
    return df, s, X, df[site_col].to_numpy()

def run_em_Gaussian(df: pl.DataFrame, s: np.ndarray, X: np.ndarray,
                    max_iter: int = 900, tol: float = 1e-5) -> Tuple[Dict, np.ndarray, np.ndarray]:
    n = len(s)
    q25, q50, q75 = np.percentile(s, [25, 50, 75])
    mask_bg = s <= q50
    mask_sig = s > q50
    mu0 = np.mean(s[mask_bg]) if np.any(mask_bg) else q25
    sigma0 = np.std(s[mask_bg]) if np.any(mask_bg) else (q50 - q25) / 2
    mu1 = np.mean(s[mask_sig]) if np.any(mask_sig) else q75
    sigma1 = np.std(s[mask_sig]) if np.any(mask_sig) else (q75 - q50) / 2
    sigma0 = max(sigma0, 0.1)
    sigma1 = max(sigma1, 0.1)

    r = np.zeros(n)
    for i in range(n):
        p_bg = stats.norm.pdf(s[i], mu0, sigma0)
        p_sig = stats.norm.pdf(s[i], mu1, sigma1)
        r[i] = p_sig / (p_bg + p_sig + 1e-10)
    r = np.clip(r, 0.01, 0.99)
    beta = irls_logistic(X, r, max_iter=20)

    log_likelihood_prev = -np.inf
    for iteration in range(max_iter):
        p = sigmoid(X @ beta)
        eps = 1e-10
        sigma0 = max(sigma0, eps)
        sigma1 = max(sigma1, eps)
        log_p_bg = stats.norm.logpdf(s, mu0, sigma0)
        log_p_sig = stats.norm.logpdf(s, mu1, sigma1)
        p_bg = np.exp(log_p_bg)
        p_sig = np.exp(log_p_sig)
        numerator = p * p_sig
        denominator = (1 - p) * p_bg + p * p_sig + eps
        r = numerator / denominator
        r = np.clip(r, eps, 1 - eps)
        try:
            beta = irls_logistic(X, r, max_iter=20)
        except:
            beta = beta * 0.95
        w0 = 1 - r
        sum_w0 = np.sum(w0) + eps
        mu0 = np.sum(w0 * s) / sum_w0
        sigma0_sq = np.sum(w0 * (s - mu0)**2) / sum_w0
        sigma0 = np.sqrt(max(sigma0_sq, eps))
        w1 = r
        sum_w1 = np.sum(w1) + eps
        mu1 = np.sum(w1 * s) / sum_w1
        sigma1_sq = np.sum(w1 * (s - mu1)**2) / sum_w1
        sigma1 = np.sqrt(max(sigma1_sq, eps))
        p_updated = sigmoid(X @ beta)
        log_p_bg_updated = stats.norm.logpdf(s, mu0, sigma0)
        log_p_sig_updated = stats.norm.logpdf(s, mu1, sigma1)
        max_log = np.maximum(
            np.log(1 - p_updated + eps) + log_p_bg_updated,
            np.log(p_updated + eps) + log_p_sig_updated
        )
        log_likelihood = np.sum(
            max_log + np.log(
                np.exp(np.log(1 - p_updated + eps) + log_p_bg_updated - max_log) +
                np.exp(np.log(p_updated + eps) + log_p_sig_updated - max_log)
            )
        )
        if abs(log_likelihood - log_likelihood_prev) < tol * (1 + abs(log_likelihood_prev)):
            break
        log_likelihood_prev = log_likelihood

    p_final = sigmoid(X @ beta)
    p_bg_final = stats.norm.pdf(s, mu0, sigma0)
    p_sig_final = stats.norm.pdf(s, mu1, sigma1)
    denominator_final = (1 - p_final) * p_bg_final + p_final * p_sig_final + 1e-10
    r_final = (p_final * p_sig_final) / denominator_final
    result = {
        "mode": "em_gaussian",
        "beta": beta.tolist(),
        "gaussian_params": {
            "mu0": float(mu0), "sigma0": float(sigma0),
            "mu1": float(mu1), "sigma1": float(sigma1)
        },
        "diagnostics": {
            "iters": iteration + 1,
            "converged": iteration < max_iter - 1,
            "final_log_likelihood": float(log_likelihood),
            "mean_p": float(p_final.mean()),
            "mean_r": float(r_final.mean())
        }
    }
    return result, p_final, r_final

def fit_model(input_parquet: str,
              out_prefix: Optional[str] = None,
              rbp_name: str = "RBP",
              motif_id: str = "Motif",
              feature_columns: Optional[List[str]] = None,
              site_col: str = "id",
              s_col: str = "s_l"):
    feature_columns = feature_columns or ['inner_mean_logPWM', 'outer_mean_logPWM', 'GC_inner_pct', 'GC_outer_pct',
                                          'TSS_proximity', 'PhastCons100_percent', 'TPM', 'score_phastcons100', 'score_phylop100']
    skip_normalization = ['TSS_proximity', 'PhastCons100_percent', 'score_phastcons100']
    df = pl.read_parquet(input_parquet)
    for c in [site_col, s_col] + feature_columns:
        if c not in df.columns:
            raise SystemExit(f"Missing column {c} in {input_parquet}")
    df_pl, s, X, _ = read_polars_input(df, site_col, s_col, feature_columns)
    X_z = X.copy()
    for i, col in enumerate(feature_columns):
        if col not in skip_normalization:
            col_z = stats.zscore(X[:, i])
            X_z[:, i] = np.nan_to_num(col_z, nan=0.0)
    X_z[:, 0] = 1
    log_s = np.log(s + 1)
    model, p, r = run_em_Gaussian(df_pl, log_s, X_z)
    out_df = df_pl.with_columns([
        pl.Series("prior_p", p),
        pl.Series("posterior_r", r),
    ])
    prefix = out_prefix or f"results/{rbp_name}_{motif_id}"
    out_parquet = f"{prefix}.model.parquet"
    out_df.write_parquet(out_parquet)
    with open(f"{prefix}.model.json", "w") as fh:
        json.dump(model, fh, indent=2)
    return out_parquet, f"{prefix}.model.json"


def build_arg_parser() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(prog="accmix model")
    ap.add_argument("--input-parquet", required=True)
    ap.add_argument("--out-prefix", default=None)
    ap.add_argument("--rbp-name", default="RBP")
    ap.add_argument("--motif-id", default="Motif")
    return ap
